import itertools
from collections import defaultdict
from pathlib import Path
from typing import List, Optional, Union

import pandas as pd
import xmltodict
from ecoinvent_interface import CachedStorage, ReleaseType
from loguru import logger
from randonneur import Datapackage, MappingConstants

from ecoinvent_migrate import __version__
from ecoinvent_migrate.data_io import get_change_report, load_release_data
from ecoinvent_migrate.ei_release import get_ei_release
from ecoinvent_migrate.patches import (
    TECHNOSPHERE_PATCHES_MISSING_DATA,
    TECHNOSPHERE_PATCHES_REPLACEMENT_DATA,
)
from ecoinvent_migrate.utils import configure_logs, setup_output_directory
from ecoinvent_migrate.wrangling import (
    apply_missing_patches,
    apply_replacement_patches,
    relabel,
    resolve_glo_row_rer_roe,
    source_target_biosphere_pair,
    source_target_pair_as_dict,
    split_replace_disaggregate,
    tuple_key_for_data,
)


def generate_technosphere_mapping(
    source_version: str,
    target_version: str,
    project_name: str = "ecoinvent-migration",
    system_model: str = "cutoff",
    ecoinvent_username: Optional[str] = None,
    ecoinvent_password: Optional[str] = None,
    write_logs: bool = True,
    write_file: bool = True,
    licenses: Optional[List[dict]] = None,
    output_directory: Optional[Path] = None,
    output_version: str = "2.0.0",
    description: Optional[str] = None,
) -> Union[Path, Datapackage]:
    """Generate a Randonneur mapping file for technosphere edge attributes from source to target."""
    configure_logs(write_logs=write_logs)

    release = get_ei_release(
        ecoinvent_username=ecoinvent_username,
        ecoinvent_password=ecoinvent_password,
    )
    source_lookup = load_release_data(
        version=source_version, system_model=system_model, release=release
    )
    target_lookup = load_release_data(
        version=target_version, system_model=system_model, release=release
    )
    excel_filepath = get_change_report(
        source_version=source_version,
        target_version=target_version,
        release=release,
    )

    sheet_names = pd.ExcelFile(excel_filepath).sheet_names
    candidates = [name for name in sheet_names if name.lower() == "qualitative changes"]
    if not candidates:
        raise ValueError(
            "Can't find suitable sheet name in change report file. Looking for 'qualitative changes', found:\n\t{}".format(
                "\n\t".join(sheet_names)
            )
        )
    elif len(candidates) > 1:
        raise ValueError(
            "Found multiple sheet names like 'qualitative changes' for change report file:\n\t{}".format(
                "\n\t".join(sheet_names)
            )
        )

    data = [
        pair
        for index, row in enumerate(
            pd.read_excel(io=excel_filepath, sheet_name=candidates[0]).to_dict(orient="records")
        )
        for pair in source_target_pair_as_dict(
            row, index + 2, excel_filepath.name, source_version, target_version
        )
    ]

    source_db_name = f"ecoinvent-{source_version}-{system_model}"
    target_db_name = f"ecoinvent-{target_version}-{system_model}"

    if not description:
        description = f"Data migration file from {source_db_name} to {target_db_name} generated with `ecoinvent_migrate` version {__version__}"

    try:
        data = apply_replacement_patches(
            data, TECHNOSPHERE_PATCHES_REPLACEMENT_DATA[(source_version, target_version)]
        )
    except KeyError:
        pass
    try:
        data = apply_missing_patches(
            data, TECHNOSPHERE_PATCHES_MISSING_DATA[(source_version, target_version)]
        )
    except KeyError:
        pass

    data = resolve_glo_row_rer_roe(
        data=data,
        source_db_name=source_db_name,
        target_db_name=target_db_name,
        source_lookup=source_lookup,
        target_lookup=target_lookup,
    )

    changed_sources = (
        set(source_lookup)
        .difference(target_lookup)
        .difference({tuple_key_for_data(line["source"]) for line in data})
    )
    for item in changed_sources:
        logger.warning(
            "Source dataset changed but neither change report nor patches have migrations: {s}",
            s=source_lookup[item],
        )

    data = [{"source": relabel(obj["source"]), "target": relabel(obj["target"])} for obj in data]
    data = split_replace_disaggregate(data=data, target_lookup=target_lookup)

    if not data["replace"] and not data["disaggregate"]:
        logger.info(
            "It seems like there are no technosphere changes for this release. Doing nothing."
        )
        return
    if not data["replace"]:
        del data["replace"]
    if not data["disaggregate"]:
        del data["disaggregate"]

    dp = Datapackage(
        name=f"{source_db_name}-{target_db_name}",
        description=description,
        contributors=[
            {
                "title": "ecoinvent association",
                "path": "https://ecoinvent.org/",
                "roles": ["author"],
            },
            {"title": "Chris Mutel", "path": "https://chris.mutel.org/", "roles": ["wrangler"]},
        ],
        mapping_source=MappingConstants.ECOSPOLD2,
        mapping_target=MappingConstants.ECOSPOLD2,
        homepage="https://github.com/brightway-lca/ecoinvent_migrate",
        version="2.0.0",
        source_id=source_db_name,
        target_id=target_db_name,
        licenses=licenses,
    )

    for key, value in data.items():
        dp.add_data(key, value)

    if write_file:
        filename = f"{source_db_name}-{target_db_name}.json"
        output_directory = setup_output_directory(output_directory)
        fp = output_directory / filename
        logger.info("Writing output file {fp}", fp=str(fp))
        return dp.to_json(fp)
    else:
        return dp


def generate_biosphere_mapping(
    source_version: str,
    target_version: str,
    keep_deletions: bool = False,
    project_name: str = "ecoinvent-migration",
    ecoinvent_username: Optional[str] = None,
    ecoinvent_password: Optional[str] = None,
    write_logs: bool = True,
    write_file: bool = True,
    licenses: Optional[List[dict]] = None,
    output_directory: Optional[Path] = None,
    output_version: str = "3.0.0",
    description: Optional[str] = None,
) -> Optional[Path]:
    """Generate a Randonneur mapping file for biosphere edge attributes from source to target."""
    configure_logs(write_logs=write_logs)

    release = get_ei_release(
        ecoinvent_username=ecoinvent_username,
        ecoinvent_password=ecoinvent_password,
    )
    load_release_data(verson=source_version, system_model="cutoff", release=release)
    load_release_data(verson=target_version, system_model="cutoff", release=release)
    excel_filepath = get_change_report(
        source_version=source_version,
        target_version=target_version,
        release=release,
    )

    source_db_name = f"ecoinvent-{source_version}-biosphere"
    target_db_name = f"ecoinvent-{target_version}-biosphere"

    logger.info(
        """The `EE Deletions` format is not consistent across versions.
Please check the outputs carefully before applying them."""
    )

    sheet_names = pd.ExcelFile(excel_filepath).sheet_names
    candidates = [name for name in sheet_names if name.lower() == "ee deletions"]
    if not candidates:
        logger.info(
            "It seems like there are no biosphere changes; no sheet name like `EE Deletions` found. Sheet names found:\n\t{sn}. Looking at actual data to see if there are changes not included in the change report.",
            sn="\n\t".join(sheet_names),
        )
        missing_sheet = True
    elif len(candidates) > 1:
        raise ValueError(
            "Found multiple sheet names like 'EE Deletions' for change report file:\n\t{}".format(
                "\n\t".join(sheet_names)
            )
        )
    else:
        missing_sheet = False

    if not description:
        description = f"Data migration file from {source_db_name} to {target_db_name} generated with `ecoinvent_migrate` version {__version__}"

    if not missing_sheet:
        # Try reading the sheet
        df = pd.read_excel(io=excel_filepath, sheet_name=candidates[0])

        # Handle the multi-index case
        if df.columns[0].startswith("**"):
            logger.debug("Detected multi-index format, adjusting reading parameters")
            df = pd.read_excel(io=excel_filepath, sheet_name=candidates[0], skiprows=1)

        # Handle the new format case
        if "deleted exchanges" in df.columns:
            logger.debug("Detected new exchange format, adjusting data structure")
            # Get the actual column headers from the first row
            new_headers = {col: val for col, val in df.iloc[0].items() if isinstance(val, str)}
            df = df.rename(columns=new_headers).iloc[1:]

        if df.empty:
            logger.info(
                "EE Deletions sheet is empty in change report for {source_v} to {target_v}. This likely means no biosphere changes.",
                source_v=source_version,
                target_v=target_version,
            )
            data = {"delete": [], "replace": []}
        else:
            data = df.to_dict(orient="records")
            data = source_target_biosphere_pair(
                data=data,
                source_version=source_version,
                target_version=target_version,
                keep_deletions=keep_deletions,
            )
            # Ensure both keys exist
            if "delete" not in data:
                data["delete"] = []
            if "replace" not in data:
                data["replace"] = []

            affected_uuids = {
                o["source"]["uuid"]
                for o in itertools.chain(data.get("replace", []), data.get("delete", []))
            }
            data = supplement_biosphere_changes_with_real_data_comparison(
                data=data,
                affected_uuids=affected_uuids,
                source_version=source_version,
                target_version=target_version,
            )
    else:
        data = supplement_biosphere_changes_with_real_data_comparison(
            data={"delete": [], "replace": []},
            affected_uuids=set(),
            source_version=source_version,
            target_version=target_version,
        )

    # Ensure we have non-empty data before creating Datapackage
    has_data = False
    cleaned_data = {}
    for key in ["delete", "replace"]:
        if data.get(key) and len(data[key]) > 0:
            cleaned_data[key] = data[key]
            has_data = True

    if not has_data:
        logger.info("No valid biosphere changes found after processing. Doing nothing.")
        return None

    dp = Datapackage(
        name=f"{source_db_name}-{target_db_name}",
        description=description,
        contributors=[
            {
                "title": "ecoinvent association",
                "path": "https://ecoinvent.org/",
                "role": "author",
            },
            {"title": "Chris Mutel", "path": "https://chris.mutel.org/", "role": "wrangler"},
        ],
        mapping_source=MappingConstants.ECOSPOLD2_BIO,
        mapping_target=MappingConstants.ECOSPOLD2_BIO,
        homepage="https://github.com/brightway-lca/ecoinvent_migrate",
        version="2.0.0",
        source_id=source_db_name,
        target_id=target_db_name,
        licenses=licenses,
    )

    # Only add non-empty data sections
    for key, value in cleaned_data.items():
        dp.add_data(key, value)

    if write_file:
        filename = f"{source_db_name}-{target_db_name}.json"
        output_directory = setup_output_directory(output_directory)
        fp = output_directory / filename
        logger.info("Writing output file {fp}", fp=str(fp))
        return dp.to_json(fp)
    else:
        return dp


def supplement_biosphere_changes_with_real_data_comparison(
    data: dict, affected_uuids: set, source_version: str, target_version: str
) -> dict:
    cs = CachedStorage()

    def format(ecospold: dict) -> dict:
        return {
            obj["@id"]: {
                "name": obj["name"]["#text"].strip(),
                "formula": obj.get("@formula").strip() if obj.get("@formula") else None,
                "unit": obj["unitName"]["#text"].strip(),
            }
            for obj in ecospold["validElementaryExchanges"]["elementaryExchange"]
        }

    source_ee = format(
        xmltodict.parse(
            open(
                Path(
                    cs.catalogue[
                        ReleaseType.ecospold.filename(
                            version=source_version, system_model_abbr="cutoff"
                        )
                    ]["path"]
                )
                / "MasterData"
                / "ElementaryExchanges.xml",
                "rb",
            )
        )
    )
    target_ee = format(
        xmltodict.parse(
            open(
                Path(
                    cs.catalogue[
                        ReleaseType.ecospold.filename(
                            version=target_version, system_model_abbr="cutoff"
                        )
                    ]["path"]
                )
                / "MasterData"
                / "ElementaryExchanges.xml",
                "rb",
            )
        )
    )

    # Patch changes which aren't included in the change report
    for key_source, value_source in source_ee.items():
        if key_source not in target_ee and key_source not in affected_uuids:
            if "delete" not in data:
                data["delete"] = []
            data["delete"].append(
                {
                    "source": {"uuid": key_source, "name": value_source["name"]},
                    "comment": "Deleted flow not listed in change report",
                }
            )
            continue
        elif key_source not in target_ee:
            # Included in change report
            continue

        diff = {
            key: value
            for key, value in target_ee[key_source].items()
            if value and value != value_source[key]
        }
        if diff:
            data["replace"].append(
                {
                    "source": {k: v for k, v in value_source.items() if v} | {"uuid": key_source},
                    "target": diff | {"uuid": key_source},
                    "comment": "Flow attribute change not listed in change report",
                }
            )

    return data



from functools import partial

SYNCOAL_PRODUCTS = [
    "1-propanol",
    "acetone, liquid",
    "ammonia, anhydrous, liquid",
    "butyl acrylate",
    "chemical, organic",
    "diesel, low-sulfur",
    "ethanol, without water, in 99.7% solution state, from ethylene",
    "ethylene",
    "heavy fuel oil",
    "isobutanol",
    "liquefied petroleum gas",
    "methanol",
    "methyl ethyl ketone",
    "natural gas, high pressure",
    "pitch",
    "propylene",
    "sulfur",
]

ACETIC_ACID_BYPRODUCTS = [
    "acetone, liquid",
    "ethyl acetate",
    "formic acid",
    "methyl acetate",
    "methyl ethyl ketone",
    "propionic acid",
]


def byproduct(
    old_name: str,
    new_name: str,
    product: str,
    geography: str,
    unit: str,
    comment: str = "By-product production missed in change report during name change",
) -> dict:
    return {
        "source": {
            "activity_name": old_name,
            "product_name": product,
            "unit": unit,
            "geography": geography,
        },
        "target": {
            "activity_name": new_name,
        },
        "comment": comment,
    }


acetic_acid_production_311 = partial(
    byproduct,
    old_name="acetic acid production, butane oxidation",
    new_name="acetic acid production, from n-butane oxidation",
    comment="3.11 change report section 7.7.3. Change report is missing by-products of new model",
    unit="kg",
)

scrap_tin_sanitary_311 = partial(
    byproduct,
    old_name="treatment of scrap tin sheet, sanitary landfill",
    new_name="treatment of waste tin sheet, sanitary landfill",
)

# Synthetic fuel production, from coal, high temperature Fisher-Tropsch operations in South Africa
# Completed changed but not well-documented in change report
syncoal_311 = partial(
    byproduct,
    old_name="synthetic fuel production, from coal, high temperature Fisher-Tropsch operations",
    unit="kg",
    geography="ZA",
    new_name="synthetic fuel production, from coal, high temperature Fischer-Tropsch operations",
)


TECHNOSPHERE_PATCHES_MISSING_DATA = {
    ("3.9.1", "3.10"): [
        {
            "source": {
                "name": "modified Solvay process, Hou's process",
                "location": "GLO",
                "reference product": "ammonium chloride",
                "unit": "kg",
            },
            "target": {
                "name": "soda ash production, dense, Hou's process",
            },
            "comment": "By-product production missed in change report during name change",
        },
        {
            "source": {
                "name": "Mannheim process",
                "location": "RER",
                "reference product": "sodium sulfate, anhydrite",
                "unit": "kg",
            },
            "target": {
                "name": "hydrochloric acid production, Mannheim process",
            },
            "comment": "By-product production missed in change report during name change",
        },
        {
            "source": {
                "name": "Mannheim process",
                "location": "RoW",
                "reference product": "sodium sulfate, anhydrite",
                "unit": "kg",
            },
            "target": {
                "name": "hydrochloric acid production, Mannheim process",
            },
            "comment": "By-product production missed in change report during name change",
        },
        {
            "source": {
                "name": "wheat production, Swiss integrated production, intensive",
                "location": "CH",
                "reference product": "straw",
                "unit": "kg",
            },
            "target": {
                "name": "wheat grain production, Swiss integrated production, intensive",
            },
            "comment": "By-product production missed in change report during name change",
        },
        {
            "source": {
                "name": "wheat production, Swiss integrated production, extensive",
                "location": "CH",
                "reference product": "straw",
                "unit": "kg",
            },
            "target": {
                "name": "wheat grain production, Swiss integrated production, extensive",
            },
            "comment": "By-product production missed in change report during name change",
        },
        {
            "source": {
                "name": "air separation, cryogenic",
                "reference product": "nitrogen, liquid",
                "location": "RER",
                "unit": "kg",
            },
            "target": {
                "name": "industrial gases production, cryogenic air separation",
            },
            "comment": "By-product production missed in change report during name change",
        },
        {
            "source": {
                "name": "air separation, cryogenic",
                "reference product": "nitrogen, liquid",
                "location": "RoW",
                "unit": "kg",
            },
            "target": {
                "name": "industrial gases production, cryogenic air separation",
            },
            "comment": "By-product production missed in change report during name change",
        },
        {
            "source": {
                "name": "air separation, cryogenic",
                "reference product": "argon, crude, liquid",
                "location": "RER",
                "unit": "kg",
            },
            "target": {
                "name": "industrial gases production, cryogenic air separation",
            },
            "comment": "By-product production missed in change report during name change",
        },
        {
            "source": {
                "name": "air separation, cryogenic",
                "reference product": "argon, crude, liquid",
                "location": "RoW",
                "unit": "kg",
            },
            "target": {
                "name": "industrial gases production, cryogenic air separation",
            },
            "comment": "By-product production missed in change report during name change",
        },
    ],
    ("3.10.1", "3.11"): (
        [syncoal_311(product=product) for product in SYNCOAL_PRODUCTS]
        + [
            {
                "source": {
                    "activity_name": "synthetic fuel production, from coal, high temperature Fisher-Tropsch operations",
                    "product_name": "1-butanol",
                    "unit": "kg",
                    "geography": "ZA",
                },
                "target": {
                    "activity_name": "synthetic fuel production, from coal, high temperature Fischer-Tropsch operations",
                    "product_name": "n-butanol",
                },
                "comment": "By-product production missed in change report during name change.",
            },
        ]
        + [
            scrap_tin_sanitary_311(product="electricity, medium voltage", unit="kWh", geography=geo)
            for geo in ("EC", "PE", "RoW")
        ]
        + [
            scrap_tin_sanitary_311(
                product="heat, district or industrial, other than natural gas",
                unit="MJ",
                geography=geo,
            )
            for geo in ("EC", "PE", "RoW")
        ]
        + [
            acetic_acid_production_311(geography=geo, product=product)
            for geo in ("RER", "RoW")
            for product in ACETIC_ACID_BYPRODUCTS
        ]
        + [
            # Natural gas transport modelling changed
            {
                "source": {
                    "activity_name": "natural gas, high pressure, import from NL",
                    "product_name": "natural gas, high pressure",
                    "unit": "m3",
                    "geography": "IT",
                },
                "target": {
                    "activity_name": "natural gas, high pressure, import from BE",
                },
                "comment": "3.11 change report section 5.3.3.4, Table 30. Hard to understand text above table 30; best-effort replacement activity.",
            },
            {
                "source": {
                    "activity_name": "natural gas, high pressure, import from NL",
                    "product_name": "natural gas, high pressure",
                    "unit": "m3",
                    "geography": "FR",
                },
                "target": {
                    "activity_name": "natural gas, high pressure, import from BE",
                },
                "comment": "3.11 change report section 5.3.3.4, Table 30. Hard to understand text above table 30; best-effort replacement activity.",
            },
            {
                "source": {
                    "activity_name": "natural gas, high pressure, import from NL",
                    "product_name": "natural gas, high pressure",
                    "unit": "m3",
                    "geography": "GB",
                },
                "target": {
                    "activity_name": "natural gas, high pressure, import from NO",
                },
                "comment": "3.11 change report section 5.3.3.4, Table 30. Hard to understand text above table 30; best-effort replacement activity.",
            },
            {
                "source": {
                    "activity_name": "natural gas, high pressure, import from FR",
                    "product_name": "natural gas, high pressure",
                    "unit": "m3",
                    "geography": "NL",
                },
                "target": {
                    "activity_name": "natural gas, high pressure, import from BE",
                },
                "comment": "3.11 change report section 5.3.3.4, Table 30. Hard to understand text above table 30; best-effort replacement activity.",
            },
            # Kraft paper update removed some by-products
            {
                "source": {
                    "activity_name": "kraft paper production",
                    "product_name": "electricity, high voltage",
                    "unit": "kWh",
                    "geography": "RER",
                },
                "target": {
                    "activity_name": "sulfate pulp production, from softwood, unbleached",
                },
                "comment": "By-product production removed and not documented in 3.11 change report (see section 14.2); Technology switch as best-effort replacement",
            },
            # Swiss crop production disaggregated by space and time missing by-products
            {
                "source": {
                    "activity_name": "wheat grain production, organic",
                    "product_name": "straw, organic",
                    "unit": "kg",
                    "geography": "CH",
                },
                "targets": [
                    {
                        f"activity_name": f"wheat grain production, {season}, organic, {region} region",
                    }
                    for season in ("spring", "winter")
                    for region in ("hill", "mountain", "plain")
                ]
                + [
                    {"activity_name": f"wheat grain, feed production, organic, {region} region"}
                    for region in ("hill", "mountain", "plain")
                ],
                "comment": "By-product production missed in change report during name change (see section 12.2)",
            },
            {
                "source": {
                    "activity_name": "barley grain production, organic",
                    "product_name": "straw, organic",
                    "unit": "kg",
                    "geography": "CH",
                },
                "targets": [
                    {
                        "activity_name": f"barley grain production, winter, organic, {region} region",
                    }
                    for region in ("hill", "mountain", "plain")
                ],
                "comment": "By-product production missed in change report during name change (see section 12.2)",
            },
            {
                "source": {
                    "activity_name": "rye production, organic",
                    "product_name": "straw, organic",
                    "unit": "kg",
                    "geography": "CH",
                },
                "targets": [
                    {
                        "activity_name": f"rye grain production, winter, organic, {region} region",
                    }
                    for region in ("hill", "mountain", "plain")
                ],
                "comment": "By-product production missed in change report during name change (see section 12.2)",
            },
            # This needs a more specialized approach as global average is 25 years out of date...
            # ] + [
            #     # Some natural gas transport datasets deleted
            #     {
            #         "source": {
            #             "activity_name": 'transport, pipeline, onshore, long distance, natural gas',
            #             "product_name": "transport, pipeline, onshore, long distance, natural gas",
            #             "unit": "metric ton*km",
            #             "geography": geo,
            #         },
            #         "target": {
            #             "activity_name": 'transport, pipeline, onshore, long distance, natural gas',
            #             "product_name": "transport, pipeline, onshore, long distance, natural gas",
            #             "unit": "metric ton*km",
            #             "geography": geo,
            #         }
            #         "comment": "By-product production missed in change report during name change (see section 12.2)",
            #     } for geo in ("AE","BR","CN","CO","DE","EC","GB","IQ","KW","MY","NG","QA","RO","SA","VE")
        ]
    ),
}


TECHNOSPHERE_PATCHES_REPLACEMENT_DATA = {
    # This is for cases where there is an entry in the change report, but that entry is incorrect.
    ("3.10.1", "3.11"): [
        # Misc. errors in change report or source data
        {
            "source": {
                "activity_name": "market for straw",
                "geography": "RER",
                "product_name": "straw",
                "unit": "kg",
            },
            "target": {
                "geography": "Europe without Switzerland",
            },
            "context": "target",
            "comment": "Data error in change report geography.",
        },
        {
            "source": {
                "activity_name": "treatment of used refrigerant R134a, reclamation",
                "product_name": "used refrigerant R134a",
                "unit": "kg",
                "geography": "GLO",
            },
            "target": {"product_name": "refrigerant R134a"},
            "comment": "Data error in original dataset; should have treated used refrigerant. Change report assumes correct product.",
            "context": "source",
        },
        # Both change report and change documentation think product is `flakes`, but it's actually
        # `pellets`
        {
            "source": {
                "activity_name": "pelletising of polyethylene, high density",
                "geography": "RER",
                "product_name": "polyethylene, high density, flakes, recycled",
                "unit": "kg",
            },
            "target": {
                "product_name": "polyethylene, high density, pellets, recycled",
            },
            "context": "target",
            "comment": "Data error in change report product name.",
        },
        {
            "source": {
                "activity_name": "pelletising of polyethylene terephthalate",
                "geography": "RER",
                "product_name": "polyethylene terephthalate, flakes, recycled",
                "unit": "kg",
            },
            "target": {
                "product_name": "polyethylene terephthalate, pellets, recycled",
            },
            "context": "target",
            "comment": "Data error in change report product name.",
        },
        {
            "source": {
                "activity_name": "pelletising of polyethylene terephthalate, food grade",
                "geography": "RER",
                "product_name": "polyethylene terephthalate, flakes, food grade, recycled",
                "unit": "kg",
            },
            "target": {
                "product_name": "polyethylene terephthalate, pellets, food grade, recycled",
            },
            "context": "target",
            "comment": "Data error in change report product name.",
        },
        # Change report uses unsorted reference product
        {
            "source": {
                "activity_name": "treatment of waste polyethylene terephthalate, for recycling, unsorted, sorting",
                "geography": "CH",
                "product_name": "waste polyethylene terephthalate, for recycling, unsorted",
                "unit": "kg",
            },
            "target": {
                "product_name": "waste polyethylene terephthalate, for recycling, sorted",
            },
            "context": "source",
            "comment": "Change report has unsorted reference product, should be sorted.",
        },
        {
            "source": {
                "activity_name": "treatment of waste polyethylene terephthalate, for recycling, unsorted, sorting",
                "geography": "Europe without Switzerland",
                "product_name": "waste polyethylene terephthalate, for recycling, unsorted",
                "unit": "kg",
            },
            "target": {
                "product_name": "waste polyethylene terephthalate, for recycling, sorted",
            },
            "context": "source",
            "comment": "Change report has unsorted reference product, should be sorted.",
        },
        {
            "source": {
                "activity_name": "treatment of waste polyethylene terephthalate, for recycling, unsorted, sorting",
                "geography": "GLO",
                "product_name": "waste polyethylene terephthalate, for recycling, unsorted",
                "unit": "kg",
            },
            "target": {
                "product_name": "waste polyethylene terephthalate, for recycling, sorted",
            },
            "context": "target",
            "comment": "Change report has unsorted reference product, should be sorted.",
        },
    ]
}



import itertools
import math
from collections import defaultdict
from copy import copy
from numbers import Number
from typing import List, Union

from loguru import logger

from ecoinvent_migrate.errors import Mismatch, Uncombinable


def isnan(o: Union[str, Number]) -> bool:
    return isinstance(o, Number) and math.isnan(o)


def tuple_key_for_data(obj: dict) -> tuple:
    if "reference product" in obj:
        return tuple([obj[field] for field in ("name", "location", "reference product", "unit")])
    else:
        return tuple(
            [obj[field] for field in ("activity_name", "geography", "product_name", "unit")]
        )


def split_by_semicolon(row: dict, version: str) -> list[dict]:
    """Turn a data `row` into one or more dictionaries.

    Splits reference product and unit values by `;\n`."""
    if isinstance(row[f"Activity Name - {version}"], Number) and math.isnan(
        row[f"Activity Name - {version}"]
    ):
        return []

    len_product = len(row[f"Reference Product - {version}"].split(";\n"))
    len_unit = len(row[f"Reference Product Unit - {version}"].split(";\n"))
    if len_product != len_unit:
        raise Mismatch(f"Can't match {len_product} products to {len_unit} units")

    return [
        {
            "activity_name": row[f"Activity Name - {version}"],
            "geography": row[f"Geography - {version}"],
            "product_name": x,
            "unit": y,
        }
        for x, y in zip(
            row[f"Reference Product - {version}"].split(";\n"),
            row[f"Reference Product Unit - {version}"].split(";\n"),
        )
    ]


def relabel(obj: dict) -> dict:
    """Change from ecospold2-ish labels to Randonneur constants.ECOSPOLD2 labels"""
    obj = copy(obj)
    obj["name"] = obj.pop("activity_name")
    obj["reference product"] = obj.pop("product_name")
    obj["location"] = obj.pop("geography")
    return obj


def source_target_pair_as_dict(
    row: dict, row_index: int, filename: str, source_version: str, target_version: str
) -> list[dict]:
    """Transform a row from the change report dataframe into one or more dicts.

    Can be more than one because the change report is for master data which can have
    multiple reference products.

    Only include the attributes we use for linking:

    * activity_name
    * product_name
    * unit
    * geography

    Example usage:

    ```python
    >>> given = {
        'Activity Name - 3.9.1': 'baling',
        'Geography - 3.9.1': 'GLO',
        'Reference Product - 3.9.1': 'baling',
        'Reference Product Unit - 3.9.1': 'unit',
        'Activity Name - 3.10': 'baling',
        'Geography - 3.10': 'GLO',
        'Reference Product - 3.10': 'baling',
        'Reference Product Unit - 3.10': 'unit',
    }
    >>> source_target_pair_as_dict(given)
    [{
        'source': {
            'activity_name': 'baling',
            'geography': 'GLO',
            'product_name': 'baling',
            'unit': 'unit'
        },
        'target': {
            'name': 'baling',
            'geography': 'GLO',
            'product_name': 'baling',
            'unit': 'unit'
        }
    }]
    ```

    But multiple pairs are returned when more than one reference product is present:

    ```python
    >>> given = {
        'Activity Name - 3.9.1': 'autoclaved aerated concrete block production',
        'Geography - 3.9.1': 'IN',
        'Reference Product - 3.9.1': 'autoclaved aerated concrete block;\nhard coal ash',
        'Reference Product Unit - 3.9.1': 'kg;\nkg',
        'Activity Name - 3.10': 'autoclaved aerated concrete block production',
        'Geography - 3.10': 'IN',
        'Reference Product - 3.10': 'autoclaved aerated concrete block;\nhard coal ash',
        'Reference Product Unit - 3.10': 'kg;\nkg',
    }
    >>> source_target_pair_as_dict(given)
    [{
        'source': {
            'activity_name': 'autoclaved aerated concrete block production',
            'geography': 'IN',
            'product_name': 'autoclaved aerated concrete block',
            'unit': 'kg'
        },
        'target': {
            'activity_name': 'autoclaved aerated concrete block production',
            'geography': 'IN',
            'product_name': 'autoclaved aerated concrete block',
            'unit': 'kg'
        }
    }, {
        'source': {
            'activity_name': 'autoclaved aerated concrete block production',
            'geography': 'IN',
            'product_name': 'hard coal ash',
            'unit': 'kg'
        },
        'target': {
            'activity_name': 'autoclaved aerated concrete block production',
            'geography': 'IN',
            'product_name': 'hard coal ash',
            'unit': 'kg'
        }
    }]
    ```

    The line breaks are in the change report, and are from individual rows which refer to multiple
    products.

    """
    versions = [x.split(" - ")[-1].strip() for x in row if x.startswith("Activity Name")]
    if f"Activity Name - {source_version}" not in row:
        raise ValueError(
            f"""Can't find source version {source_version} in data row.
    Versions found: {versions}"""
        )
    if f"Activity Name - {target_version}" not in row:
        raise ValueError(
            f"""Can't find target version {target_version} in data row.
    Versions found: {versions}"""
        )

    sources = split_by_semicolon(row, source_version)
    if not sources:
        # New unit process dataset, no source objects
        return []

    targets = split_by_semicolon(row, target_version)
    if len(sources) > 1 and len(targets) > 1 and len(sources) != len(targets):
        raise Uncombinable(
            f"Can't do M:N combination of {len(sources)} source and {len(targets)} target datasets."
        )
    elif len(sources) == 1 and len(targets) == 1:
        return [
            {
                "source": sources[0],
                "target": targets[0],
                "comment": f"Line {row_index} in change report file `{filename}`",
            }
        ]
    elif len(sources) == 1:
        sources = itertools.repeat(sources[0])
    elif len(targets) == 1:
        targets = itertools.repeat(targets[0])

    return [
        {
            "source": s,
            "target": t,
            "comment": f"Line {row_index} in change report file `{filename}`",
        }
        for s, t in zip(sources, targets)
        if all(v.lower() != "nan" for v in itertools.chain(s.values(), t.values()))
    ]


def resolve_glo_row_rer_roe(
    data: List[dict],
    source_db_name: str,
    target_db_name: str,
    source_lookup: dict,
    target_lookup: dict,
) -> List[dict]:
    """Iterate through `data`, and change `geography` attribute to `RoW` or `RoE` when needed.

    Looks in actual database to get correct `geography` attributes."""

    warned = set()

    for obj in data:
        source_missing = None
        for kind, lookup, db_name in [
            ("source", source_lookup, source_db_name),
            ("target", target_lookup, target_db_name),
        ]:
            key = tuple_key_for_data(obj[kind])
            if key in lookup:
                continue
            elif (
                key not in lookup
                and obj[kind]["geography"] == "GLO"
                and (key[0], "RoW", key[2], key[3]) in lookup
            ):
                obj[kind]["geography"] = "RoW"
                logger.debug(
                    "{kind} process {name} geography corrected to 'RoW'",
                    kind=kind,
                    name=obj[kind]["activity_name"],
                )
            elif (
                key not in lookup
                and obj[kind]["geography"] == "RER"
                and (key[0], "RoE", key[2], key[3]) in lookup
            ):
                obj[kind]["geography"] = "RoE"
                logger.debug(
                    "{kind} process {name} geography corrected to 'RoE'",
                    kind=kind,
                    name=obj[kind]["activity_name"],
                )
            else:
                if kind == "target" and source_missing:
                    # Missing in both source and target for this system model
                    source_missing = None
                    continue
                elif kind == "source":
                    source_missing = obj[kind]
                else:
                    # Only missing in target database - but this is a big problem, we don't have a
                    # suitable target for existing edges to relink to.
                    if key := tuple_key_for_data(obj[kind]) not in warned:
                        warned.add(key)
                        logger.warning(
                            "{kind} process given in change report but missing in {db_name} lookup: {ds}",
                            kind=kind.title(),
                            db_name=db_name,
                            ds=obj[kind],
                        )
        if source_missing:
            # Only a debug message because this won't break anything - there is no process in the
            # source database to miss a link from.
            logger.debug(
                "Source process given in change report but missing in {db_name} lookup: {ds}",
                db_name=source_db_name,
                ds=source_missing,
            )

    return data


def disaggregated(data: List[dict], lookup: dict) -> dict:
    """Take a list of mapping dictionaries with the same `source`, and create one `disaggregate`
    object.

    Applies `allocation` factors based on the production volumes in `lookup`

    """
    for obj in data:
        try:
            obj["pv"] = lookup[tuple_key_for_data(obj["target"])]["production_volume"]
        except KeyError:
            logger.warning(
                """Change report annex dataset missing from database.
    This is likely a publication error which you can't fix.
    Removing the following from migrations:
    {ds}""",
                ds=obj["target"],
            )
            obj["pv"] = 0

    total = sum(obj["pv"] for obj in data)
    if not total:
        logger.warning(
            "Total production from {n} targets is zero for source {s}; using equal allocation factors",
            n=len(data),
            s=data[0]["source"],
        )
        return {
            "source": data[0]["source"],
            "targets": [obj["target"] | {"allocation": 1 / len(data)} for ob in data],
        }
    elif total < 0:
        logger.warning(
            "Total production from {n} targets is less than zero for source {s}; what is happening!?",
            n=len(data),
            s=data[0]["source"],
        )

    return {
        "source": data[0]["source"],
        "targets": [obj["target"] | {"allocation": obj["pv"] / total} for obj in data if obj["pv"]],
    }


def split_replace_disaggregate(data: List[dict], target_lookup: dict) -> dict:
    """Split the transformations in `data` into `replace` and `disaggregate` sections.

    Disaggregation is needed when one dataset is replaced by multiple datasets. We lookup the
    respective production volumes to get the disaggregation factors."""
    groupie = defaultdict(list)
    for obj in data:
        groupie[tuple_key_for_data(obj["source"])].append(obj)

    return {
        "replace": [
            value[0]
            for value in groupie.values()
            if len(value) == 1 and value[0]["source"] != value[0]["target"]
        ],
        "disaggregate": [
            disaggregated(value, target_lookup) for value in groupie.values() if len(value) > 1
        ],
    }


def get_column_labels(example: dict, version: str) -> dict:
    """Guess column labels from Excel change report annex.

    Now handles multiple formats:
    - Standard format: "UUID/ID - version"
    - New format: Where labels might be in values
    """
    uuid_tries = [f"UUID - {version}", f"ID - {version}"]
    name_tries = [f"Name - {version}", f"{version} name", f"{version} - name"]

    # Try standard format first
    for uuid_try in uuid_tries:
        if uuid_try in example:
            uuid = uuid_try
            break
    else:
        # If standard format fails, check for new format
        for key, value in example.items():
            if isinstance(value, str) and any(try_pattern in value for try_pattern in uuid_tries):
                uuid = key
                break
        else:
            raise ValueError(f"Can't find uuid field for database version {version} in {example}")

    # Same pattern for name
    for name_try in name_tries:
        if name_try in example:
            name = name_try
            break
    else:
        for key, value in example.items():
            if isinstance(value, str) and any(try_pattern in value for try_pattern in name_tries):
                name = key
                break
        else:
            raise ValueError(f"Can't find name field for database version {version} in {example}")

    return {
        "uuid": uuid,
        "name": name,
    }


def source_target_biosphere_pair(
    data: List[dict], source_version: str, target_version: str, keep_deletions: bool
) -> List[dict]:
    """Turn pandas DataFrame rows into source/target pairs.

    The function now handles both old and new EE Deletions formats:
    - Old format: Direct source/target columns
    - New format: Deletion/replacement columns with explicit relationships
    """
    # For empty data, return empty structure
    if not data:
        return {"replace": [], "delete": []}

    # Try old format first
    source_labels = get_column_labels(example=data[0], version=source_version)
    target_labels = get_column_labels(example=data[0], version=target_version)

    # Initialize the result structure
    formatted = {"replace": [], "delete": [] if keep_deletions else None}

    # Process each row
    for row in data:
        # Skip empty or invalid rows
        if any(isnan(row.get(v)) for v in source_labels.values()):
            continue

        # Create source entry
        source_entry = {k: row[v].strip() for k, v in source_labels.items()}

        # Check if there's a valid target
        has_target = not any(isnan(row.get(v, float("nan"))) for v in target_labels.values())

        if has_target:
            formatted["replace"].append(
                {
                    "source": source_entry,
                    "target": {k: row[v].strip() for k, v in target_labels.items()},
                    "conversion_factor": float(row.get("Conversion Factor (old-new)", 1.0)),
                    "comment": row.get("Comment").strip(),
                }
            )
        elif keep_deletions:
            formatted["delete"].append(
                {
                    "source": source_entry,
                    "comment": row.get("Comment").strip(),
                }
            )

    # Clean up the formatted data
    for lst in formatted.values():
        if lst is None:
            continue
        for obj in lst:
            if "comment" in obj and (not obj["comment"] or isnan(obj["comment"])):
                del obj["comment"]
            if "conversion_factor" in obj and (
                obj["conversion_factor"] == 1.0 or isnan(obj["conversion_factor"])
            ):
                del obj["conversion_factor"]

    # Remove empty delete list if not keeping deletions
    if not keep_deletions:
        del formatted["delete"]

    return formatted


def apply_missing_patches(data: list[dict], patches: list[dict]) -> list[dict]:
    """
    Apply patches which add transformations missing from the change report.

    Patches can be 1-to-1 or disaggregation. Both versions take a complete `source` dictionary, and
    use the `target` dictionary to model the changes needed to make a successful migration. Normally
    the `target` dictionary only has the elements which change.

    1-to-1 patches look like this:

    ```python
    >>> patches = [{
        "source": {
            "activity_name": "kraft paper production",
            "product_name": "electricity, high voltage",
            "unit": "kWh",
            "geography": "RER",
        },
        "target": {
            "activity_name": "sulfate pulp production, from softwood, unbleached",
        },
        "comment": "By-product production removed and not documented in 3.11 change report.",
    }]
    >>> apply_missing_patches([], patches)
    [{
        "source": {
            "activity_name": "kraft paper production",
            "product_name": "electricity, high voltage",
            "unit": "kWh",
            "geography": "RER",
        },
        "target": {
            "activity_name": "sulfate pulp production, from softwood, unbleached",
            "product_name": "electricity, high voltage",
            "unit": "kWh",
            "geography": "RER",
        },
        "comment": "By-product production removed and not documented in 3.11 change report.",
    }]
    ```

    In disaggregation, instead of a `target` dictionary, there is a `targets` dictionary with
    multiple objects. The production volume of the targets will get added later with the
    `disaggregated` function. Disaggregation patches look like this:

    ```python
    >>> patches = [{
        "source": {
            "activity_name": "wheat grain production, organic",
            "product_name": "straw, organic",
            "unit": "kg",
            "geography": "CH",
        },
        "targets": [
            {
                "activity_name": "wheat grain production, spring, organic, hill region",
            },
            {
                "activity_name": "wheat grain production, spring, organic, mountain region",
            }
        ],
        "comment": "Split of wheat production into many specific processes",
    }]
    >>> apply_missing_patches([], patches)
    [{
        "source": {
            "activity_name": "wheat grain production, organic",
            "product_name": "straw, organic",
            "unit": "kg",
            "geography": "CH",
        },
        "target": {
            "activity_name": "wheat grain production, spring, organic, hill region",
            "product_name": "straw, organic",
            "unit": "kg",
            "geography": "CH",
        },
        "comment": "Split of wheat production into many specific processes",
    }, {
        "source": {
            "activity_name": "wheat grain production, organic",
            "product_name": "straw, organic",
            "unit": "kg",
            "geography": "CH",
        },
        "target": {
            "activity_name": "wheat grain production, spring, organic, mountain region",
            "product_name": "straw, organic",
            "unit": "kg",
            "geography": "CH",
        },
        "comment": "Split of wheat production into many specific processes",
    }]
    ```

    """
    for patch in patches:
        if "targets" in patch:
            for target in patch["targets"]:
                data.append(
                    {key: value for key, value in patch.items() if key != "targets"}
                    | {"target": copy(patch["source"]) | target}
                )
        else:
            data.append(
                {key: value for key, value in patch.items() if key != "target"}
                | {"target": copy(patch["source"]) | patch["target"]}
            )

    return data


def apply_replacement_patches(data: list[dict], patches: list[dict]) -> list[dict]:
    """
    Replace elements of existing `source` or `target` dictionaries in `data`.

    Uses `context` to determine which mapping dictionary to modify.
    """
    # These will be modified during the following loop, but we assume that the patches are
    # well-behaved and don't overlap.
    source_lookup = defaultdict(list)
    target_lookup = defaultdict(list)
    for obj in data:
        source_lookup[tuple_key_for_data(obj["source"])].append(obj)
        target_lookup[tuple_key_for_data(obj["target"])].append(obj)

    for patch in patches:
        patch_key = tuple_key_for_data(patch["source"])
        for kind, lookup in (("source", source_lookup), ("target", target_lookup)):
            if patch["context"] == kind:
                if patch_key not in lookup:
                    logger.warning(
                        "Expected to patch the following {k} dict but it's not in the given data: {p}",
                        k=kind,
                        p=patch["source"],
                    )
                    continue
                for obj in lookup[patch_key]:
                    logger.debug(
                        "Patching change report {k} {s} with updated values {t}",
                        k=kind,
                        s=obj,
                        t=patch["target"],
                    )
                    obj[kind].update(**patch["target"])
                    if "comment" in patch:
                        if "comment" in obj:
                            string = (
                                ("." if not obj["comment"].endswith(".") else "")
                                + f" Patched with comment '"
                                + patch["comment"]
                                + "'."
                            )
                            obj["comment"] += string
                        else:
                            obj["comment"] = patch["comment"]
    return data



import warnings
from pathlib import Path

from ecoinvent_interface import EcoinventRelease, Settings


def get_ei_release(
    ecoinvent_username: str | None = None,
    ecoinvent_password: str | None = None,
) -> Path:
    if ecoinvent_username is not None or ecoinvent_password is not None:
        warnings.warn(
            """
Username and/or password supplied via function call.
Please consider setting these permanently. See:
https://github.com/brightway-lca/ecoinvent_interface?tab=readme-ov-file#authentication-via-settings-object
            """
        )
        settings = Settings(username=ecoinvent_username, password=ecoinvent_password)
    else:
        settings = Settings()
    return EcoinventRelease(settings)



import json
from dataclasses import asdict, dataclass
from pathlib import Path

from ecoinvent_interface import EcoinventRelease, ReleaseType
from loguru import logger
from lxml import objectify
from tqdm import tqdm

from ecoinvent_migrate.errors import VersionJump
from ecoinvent_migrate.utils import cache_dir
from ecoinvent_migrate.wrangling import tuple_key_for_data


@dataclass
class SOUPInfo:
    """
    Single-output unit process

    Has unique combination of activity name, product name, geography, and unit.
    """

    activity_name: str
    product_name: str
    unit: str
    geography: str
    production_volume: float
    filename: str

    def as_tuple(self) -> tuple[str]:
        return tuple_key_for_data(asdict(self))


def soupinfo_for_file(fp: Path) -> SOUPInfo:
    root = objectify.parse(open(fp)).getroot()
    if hasattr(root, "activityDataset"):
        stem = root.activityDataset
    else:
        stem = root.childActivityDataset

    for prod_exc in stem.flowData.iterchildren():
        if hasattr(prod_exc, "outputGroup") and prod_exc.outputGroup.text == "0":
            break
    else:
        raise ValueError("Can't find production exchange")

    return SOUPInfo(
        activity_name=stem.activityDescription.activity.activityName.text.strip(),
        product_name=prod_exc.name.text.strip(),
        unit=prod_exc.unitName.text.strip(),
        geography=stem.activityDescription.geography.shortname.text.strip(),
        production_volume=float(prod_exc.get("productionVolumeAmount") or 0),
        filename=fp.name,
    )


def load_release_data(version: str, system_model: str, release: EcoinventRelease) -> None:
    cache_filepath = cache_dir() / f"ecoinvent-{version}-{system_model}.json"
    if cache_filepath.is_file():
        return {tuple_key_for_data(obj): obj for obj in json.load(open(cache_filepath))}
    else:
        logger.info(
            "Downloading ecoinvent version {version} {system_model}",
            version=version,
            system_model=system_model,
        )
        release.get_release(version, system_model, ReleaseType.ecospold)

        dirpath = release.get_release(version, system_model, ReleaseType.ecospold) / "datasets"
        assert dirpath.is_dir(), f"Release cache at {dirpath.parent} missing `datasets` directory"

        data = [
            asdict(soupinfo_for_file(fp))
            for fp in tqdm(
                filter(lambda x: x.suffix.lower() in {".xml", ".spold"}, dirpath.iterdir())
            )
        ]
        with open(cache_filepath, "w") as f:
            json.dump(data, f, indent=2, ensure_ascii=False)

        return {tuple_key_for_data(obj): obj for obj in data}


def get_change_report(
    source_version: str,
    target_version: str,
    release: EcoinventRelease,
) -> Path:
    """Get the source/target change report filepath, and setup Brightway project with needed data.

    Source and target must be one release apart from each other, i.e. 3.4 to 3.5 or 3.7 to 3.7.1.

    Returns the `Path` of the generated file."""
    versions = release.list_versions()

    if source_version not in versions:
        raise ValueError(
            f"Given source version {source_version} not in available versions: {versions}"
        )
    if target_version not in versions:
        raise ValueError(
            f"Given target version {source_version} not in available versions: {versions}"
        )

    logger.info("Versions available for this license: {versions}", versions=versions)

    excel_filepath = get_change_report_filepath(version=target_version, release=release)
    if source_version not in excel_filepath.name.replace(target_version, ""):
        raise VersionJump(
            f"""
Source ({source_version}) and target ({target_version}) don't have a change report.
Usually this is the case when one jumps across multiple releases, but not always.
For example, the change report for 3.7.1 is from 3.6, not 3.7.
The change report we have available is:
{excel_filepath.name}
        """
        )
    logger.info("Using change report annex file {filename}", filename=excel_filepath.name)
    return excel_filepath


def get_change_report_filepath(version: str, release: EcoinventRelease) -> Path:
    """Get the filepath to the Excel change report file.

    Download a list of extra files from ecoinvent and do pattern matching."""
    files = release.list_extra_files(version)
    candidates = [key for key in files if "change report" in key.lower() and "annex" in key.lower()]
    if not candidates:
        raise ValueError(
            "Can't find suitable change report filename from release files:\n\t{}".format(
                "\n\t".join([f.name for f in files])
            )
        )
    elif len(candidates) > 1:
        raise ValueError(
            "Found multiple candidates for change report filename:\n\t{}".format(
                "\n\t".join([c.name for c in candidates])
            )
        )

    files = list(release.get_extra(version, candidates[0]).iterdir())
    candidates = [
        fp
        for fp in files
        if (
            fp.name.lower().startswith("change report annex")
            or fp.name.lower().startswith("change_report_annex")
        )
        and fp.suffix.lower() == ".xlsx"
    ]
    if not candidates:
        raise ValueError(
            "Can't find suitable change report file from archive:\n\t{}".format(
                "\n\t".join([f.name for f in files])
            )
        )
    elif len(candidates) > 1:
        raise ValueError(
            "Found multiple candidates for change report file:\n\t{}".format(
                "\n\t".join([c.name for c in candidates])
            )
        )
    return candidates[0]



import datetime
import os
import sys
from pathlib import Path
from typing import Optional

from loguru import logger
from platformdirs import user_data_dir, user_log_dir


def configure_logs(write_logs: bool = True) -> None:
    logger.remove()
    logger.add(sys.stderr, level="INFO")
    if write_logs:
        logs_dir = Path(
            user_log_dir("ecoinvent_migrate", "pylca")
        ) / datetime.datetime.now().isoformat()[:19].replace(":", "-")
        logger.info("Writing logs to {path}", path=logs_dir)
        logs_dir.mkdir(parents=True, exist_ok=True)
        logger.add(logs_dir / "debug.log", level="DEBUG")
        logger.add(logs_dir / "info.log", level="INFO")


def setup_output_directory(output_directory: Optional[Path]) -> Path:
    if output_directory is None:
        output_directory = Path(user_data_dir("ecoinvent_migrate", "pylca"))
        if not output_directory.exists():
            logger.info(f"Creating output directory {output_directory}")
            output_directory.mkdir(parents=True)
    elif not isinstance(output_directory, Path):
        raise ValueError(
            f"`output_directory` must be a `Path` instance; got {type(output_directory)}"
        )
    elif not output_directory.is_dir() or not os.access(output_directory, os.W_OK):
        raise ValueError(f"`output_directory` {output_directory} must be a writable directory")
    return output_directory


def cache_dir() -> Path:
    cache_directory = Path(user_data_dir("ecoinvent_migrate", "pylca")) / "cache"
    if not cache_directory.exists():
        cache_directory.mkdir(parents=True)
    return cache_directory
